import os
import sys
import unittest
from pypdn.nrbf import NRBF

# TODO Write NRBF unit tests
class TestNRBF(unittest.TestCase):
    def setUp(self):
        pass

    def test_read_nrbf(self):
        pass


if __name__ == '__main__':
    unittest.main()
